package me.yuxiaoyao.config.loader;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;

import java.io.*;
import java.util.*;

/**
 * @author Kerry on 19/10/09
 */

public class ConfigLoader {

    private static final String PROFILE = "profile";
    private static final String ACTIVE = "active";

    public static final String ACTIVE_PROFILE = PROFILE + "." + ACTIVE;

    /**
     * 如果想修改,可以通过反射修改
     */
    public static String DEFAULT_FILE_CONFIG_FOLDER = "config";
    private static String DEFAULT_APPLICATION = "application";

    private static String START_PLACEHOLDER = "${";
    private static String END_PLACEHOLDER = "}";


    private enum ConfigProtocol {
        /**
         * 文件
         */
        FILE(1),
        /**
         * 类路径
         */
        CLASSPATH(0);

        /**
         * 配置文件优先级
         */
        private int order;

        ConfigProtocol(int order) {
            this.order = order;
        }

        public static List<ConfigProtocol> sortList() {
            List<ConfigProtocol> configProtocols = new ArrayList<>(Arrays.asList(values()));
            configProtocols.sort((o1, o2) -> o2.order - o1.order);
            return configProtocols;
        }
    }

    public interface ConfigListener {
        /**
         * 回调Profile
         *
         * @param profile
         */
        void activeProfile(String profile);
    }

    /**
     * 返回数据流
     *
     * @param configProtocol
     * @param path
     * @return 结果可能为 null
     */
    private static InputStream getConfigFileAsStream(ConfigProtocol configProtocol, String path) {

        switch (configProtocol) {
            case FILE:
                try {
                    return new FileInputStream(new File(DEFAULT_FILE_CONFIG_FOLDER + File.separator + path));
                } catch (FileNotFoundException ignored) {
                }
                return null;
            case CLASSPATH:
                return Thread.currentThread().getContextClassLoader().getResourceAsStream(path);
            default:
                throw new Error("not support this path = " + path);
        }
    }


    public static <T> T parseConfig(Class<T> clazz) {
        return parseConfig(null, clazz, null);
    }

    public static <T> T parseConfig(String[] args, Class<T> clazz, ConfigListener configListener) {
        LinkedHashMap map = loaderConfig(args, configListener);
        return getObjectMapper().convertValue(map, clazz);
    }


    public static <T> T parseConfig(TypeReference<T> toValueTypeRef) {
        return parseConfig(null, toValueTypeRef, null);
    }

    public static <T> T parseConfig(String[] args, TypeReference<T> toValueTypeRef, ConfigListener configListener) {
        LinkedHashMap map = loaderConfig(args, configListener);
        return getObjectMapper().convertValue(map, toValueTypeRef);
    }

    private static ObjectMapper getObjectMapper() {
        ObjectMapper objectMapper = new ObjectMapper();
        objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        return objectMapper;
    }

    /**
     * 加载默认配置文件
     *
     * @return
     */
    public static LinkedHashMap loaderConfig() {
        return loaderConfig(null, null);
    }

    /**
     * 加载指定 profile 的配置文件
     *
     * @param args
     * @return
     */
    public static LinkedHashMap loaderConfig(String[] args, ConfigListener configListener) {

        List<ConfigProtocol> configProtocols = ConfigProtocol.sortList();
        LinkedHashMap defaultProperties = null;
        ConfigProtocol matchProtocol = null;
        for (ConfigProtocol protocol : configProtocols) {
            LinkedHashMap temp = selectConfig(protocol, DEFAULT_APPLICATION);
            if (temp != null) {
                defaultProperties = temp;
                matchProtocol = protocol;
                break;
            }
        }
        if (defaultProperties == null) {
            return null;
        }

        String profile = getProfileByArgs(args);

        if (profile == null || profile.length() == 0) {
            profile = getActiveProfile(defaultProperties);
        }
        if (configListener != null) {
            configListener.activeProfile(profile);
        }
        LinkedHashMap map = selectConfig(matchProtocol, DEFAULT_APPLICATION + "-" + profile);
        if (map != null) {
            removeActiveProfile(map);
            //noinspection unchecked
            defaultProperties.putAll(map);
        }
        return propertiesHandler(defaultProperties);
    }


    private static LinkedHashMap selectConfig(ConfigProtocol configProtocol, String filePrefix) {
        InputStream configFileAsStream = getConfigFileAsStream(configProtocol, filePrefix + ".yml");
        if (configFileAsStream == null) {
            configFileAsStream = getConfigFileAsStream(configProtocol, filePrefix + ".yaml");
        }
        Object yamlFactory = null;
        try {
            Class<?> factory = Class.forName("com.fasterxml.jackson.dataformat.yaml.YAMLFactory");
            yamlFactory = factory.newInstance();
        } catch (InstantiationException | IllegalAccessException | ClassNotFoundException e) {
            // e.printStackTrace();
        }
        ObjectMapper objectMapper = null;
        if (configFileAsStream != null && yamlFactory != null) {
            objectMapper = new ObjectMapper((JsonFactory) yamlFactory);
        } else {
            configFileAsStream = getConfigFileAsStream(configProtocol, filePrefix + ".properties");
            try {
                Class<?> javaProps = Class.forName("com.fasterxml.jackson.dataformat.javaprop.JavaPropsMapper");
                objectMapper = (ObjectMapper) javaProps.newInstance();
            } catch (ClassNotFoundException | IllegalAccessException | InstantiationException e) {
                // e.printStackTrace();
            }
        }
        if (objectMapper == null) {
            throw new Error("jackson dependencies at least container [jackson-dataformat-yaml] or [jackson-dataformat-properties].");
        }
        if (configFileAsStream == null) {
            return null;
        }
        try {
            return objectMapper.readValue(configFileAsStream, LinkedHashMap.class);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }


    private static String getActiveProfile(Map fileMap) {
        Object profile = fileMap.get(PROFILE);
        if (profile instanceof Map) {
            final Object o = ((Map) profile).get(ACTIVE);
            if (o != null) {
                return (String) o;
            }
        }
        return null;
    }

    private static void removeActiveProfile(Map map) {
        Object profile = map.get(PROFILE);
        if (profile instanceof Map) {
            ((Map) profile).remove(ACTIVE);
        }
    }

    /**
     * 属性处理,目前暂时不处理,替换符号等操作
     *
     * @param map
     * @return
     */
    private static LinkedHashMap propertiesHandler(LinkedHashMap map) {
        Map<Object, String> preHandler = new LinkedHashMap<>();
        // find placeholder value
        findPreResolveProperties(map, preHandler, null);
        // replace value
        Map<Object, Object> objectObjectMap = ConfigLoader.resolveProperties(map, preHandler);
        // replace to map
        replaceProperties(map, objectObjectMap);
        return map;
    }


    private static Object getProperties(Map map, String key) {
        final int i = key.indexOf(".");
        if (i == -1) {
            return map.get(key);
        }
        String currentKey = key.substring(0, i);
        Object o = map.get(currentKey);

        if (o instanceof Map) {
            return getProperties((Map) o, (key.substring(i + 1)));
        }
        return o;
    }

    @SuppressWarnings("unchecked")
    private static void setProperties(Map map, String key, Object value) {
        if (map == null) {
            return;
        }
        final int i = key.indexOf(".");
        if (i == -1) {
            map.put(key, value);
            return;
        }
        String currentKey = key.substring(0, i);
        Object o = map.get(currentKey);
        if (o instanceof Map) {
            setProperties((Map) o, (key.substring(i + 1)), value);
        } else {
            map.put(key, value);
        }
    }


    public static void replaceProperties(Map map, Map<Object, ?> handlerMap) {

        for (Object o : handlerMap.keySet()) {
            Object value = handlerMap.get(o);
            setProperties(map, o.toString(), value);

        }

    }


    /**
     * 把带的占位符的值替换为真实值
     *
     * @param map
     * @param preHandlerMap
     * @return
     */
    public static Map<Object, Object> resolveProperties(Map map, Map<Object, ?> preHandlerMap) {
        Map<Object, Object> processedMap = new LinkedHashMap<>(preHandlerMap.size());

        for (Object o : preHandlerMap.keySet()) {
            Object value = preHandlerMap.get(o);

            if (value instanceof String) {
                // source value
                String str = ((String) value).trim();
                str = str.substring(START_PLACEHOLDER.length(), str.length() - END_PLACEHOLDER.length());
                String[] defaultValue = str.split(":");
                Object properties;
                if (defaultValue.length > 1) {
                    properties = getProperties(map, defaultValue[0]);
                } else {
                    properties = getProperties(map, defaultValue[0]);
                }
                // set default value
                if (properties == null) {
                    if (defaultValue.length > 1) {
                        properties = defaultValue[1];
                    } else {
                        properties = value;
                    }
                }
                /// System.out.println("key = [" + value + "] - value = [" + properties + "]");
                processedMap.put(o, properties);
            }
        }
        return processedMap;
    }

    /**
     * 查找所有具有占位符号的值,目前不支持数组
     * 起始符 {@link ConfigLoader#START_PLACEHOLDER}
     * 结束符 {@link ConfigLoader#END_PLACEHOLDER}
     *
     * @param map
     * @param result 查找结果接收
     * @return
     */
    public static void findPreResolveProperties(Map map, Map<Object, String> result, String prefix) {

        for (Object key : map.keySet()) {
            Object value = map.get(key);
            if (value instanceof Map) {
                findPreResolveProperties((Map) value, result, prefix == null ? key.toString() : prefix + "." + key.toString());
            } else if (value instanceof String) {
                String str = ((String) value).trim();
                if (str.startsWith(START_PLACEHOLDER) && str.endsWith(END_PLACEHOLDER)) {
                    if (prefix == null) {
                        result.put(key.toString(), str);
                    } else {
                        result.put(prefix + "." + key.toString(), str);
                    }
                }
            }
        }
        //prefix = "";
    }


    private static String getProfileByArgs(String[] args) {
        if (args != null && args.length > 0) {
            for (String arg : args) {
                arg = arg.trim();
                if (arg.startsWith("--profile=")) {
                    return arg.substring("--profile=".length());
                }
            }
        }
        return null;
    }


}
